package com.emarcotte.testrio;

import java.util.ArrayList;

import javax.media.opengl.GL;

import com.emarcotte.glfoo.math.Vector3;

/**
 * Mesh coating over the joints.
 * 
 * @author Eugene Marcotte
 */

public class MD5Mesh {
	private ArrayList<MD5Weight> weights;
	private ArrayList<MD5Vertex> verts;
	private Vector3[] vertexPositions;
	private Vector3[] bindPositionNormals;
	private double[] vertsData;
	private double[] normalData;
	private double[] uvData;
	private int[] faceData;

	/**
	 * Setup the weights array, with a given number of elements.
	 * 
	 * @param numWeights
	 *            The number of weights to make storage for.
	 */
	public void setNumWeights(int numWeights) {
		weights = new ArrayList<MD5Weight>(numWeights);
	}

	/**
	 * Setup the vertices array, with a given number of elements.
	 * 
	 * @param numVertices
	 *            The number of verticies
	 */
	public void setNumVertices(int numVertices) {
		verts = new ArrayList<MD5Vertex>(numVertices);
		vertsData = new double[numVertices * 3];
		normalData = new double[numVertices * 3];
		bindPositionNormals = new Vector3[numVertices];
		vertexPositions = new Vector3[numVertices];
		uvData = new double[numVertices * 2];

		for (int i = 0; i < numVertices; i++) {
			bindPositionNormals[i] = new Vector3();
			vertexPositions[i] = new Vector3();
		}
	}

	/**
	 * Setup the triangles array, with a given number of elements.
	 * 
	 * @param numTriangles
	 *            The number of triangles to make storage for.
	 */
	public void setNumTriangles(int numTriangles) {
		faceData = new int[numTriangles * 3];
	}

	public void addVertex(int vertId, double textureU, double textureV,
			int weightIndex, int weightElem) {
		MD5Vertex v = new MD5Vertex(vertId, weightIndex, weightElem);
		verts.add(vertId, v);
		uvData[vertId * 2] = textureU;
		uvData[vertId * 2 + 1] = textureV;
	}

	/**
	 * Add a triangle to the mesh.
	 * 
	 * @param index
	 *            The ID of the triangle.
	 * @param t
	 *            The Triangle to add.
	 */
	public void addTriangle(int index, int vert1, int vert2, int vert3) {
		faceData[index * 3] = vert1;
		faceData[index * 3 + 1] = vert2;
		faceData[index * 3 + 2] = vert3;
	}

	public void addWeight(int index, double x, double y, double z,
			double weightValue, int jointIndex) {
		MD5Weight w = new MD5Weight(x, y, z, weightValue, jointIndex);
		weights.add(index, w);
	}

	public ArrayList<MD5Vertex> getVerts() {
		return verts;
	}

	public ArrayList<MD5Weight> getWeights() {
		return weights;
	}

	public void drawNormals(GL gl) {
		gl.glBegin(GL.GL_LINES);
		for (int i = 0; i < verts.size(); i++) {
			gl.glVertex3d(getX(i), getY(i), getZ(i));
			gl.glVertex3d(getX(i) + getNX(i), getY(i) + getNY(i), getZ(i)
					+ getNZ(i));
		}
		gl.glEnd();
	}

	public void computeNormals(MD5Frame skel) {
		Vector3 normal = new Vector3();

		for (int i = 0; i < bindPositionNormals.length; i++) {
			bindPositionNormals[i].zero();
		}

		// determine each faces' contribution to vertices normals.
		for (int i = 0; i < faceData.length; i += 3) {
			Vector3 v1 = vertexPositions[faceData[i]];
			Vector3 v2 = vertexPositions[faceData[i + 1]];
			Vector3 v3 = vertexPositions[faceData[i + 2]];

			// figure out face normal
			normal = computeFaceNormal(v2, v1, v3);

			bindPositionNormals[faceData[i]].addMe(normal);
			bindPositionNormals[faceData[i + 1]].addMe(normal);
			bindPositionNormals[faceData[i + 2]].addMe(normal);
		}

		// Average all the normal values for each vertex.
		for (int i = 0; i < bindPositionNormals.length; i++) {
			bindPositionNormals[i].normalize();
			// set initial normal data
			normalData[i * 3] = bindPositionNormals[i].i;
			normalData[i * 3 + 1] = bindPositionNormals[i].j;
			normalData[i * 3 + 2] = bindPositionNormals[i].k;
		}

	}

	private Vector3 computeFaceNormal(Vector3 p1, Vector3 p2, Vector3 p3) {
		Vector3 v1 = p1.subtract(p2);
		Vector3 v2 = p1.subtract(p3);
		Vector3 result = v1.crossProduct(v2);
		result.normalize();
		return result;
	}

	public double getU(int id) {
		return uvData[id * 2];
	}

	public double getV(int id) {
		return uvData[id * 2 + 1];
	}

	public double getX(int id) {
		return vertsData[id * 3];
	}

	public double getY(int id) {
		return vertsData[id * 3 + 1];
	}

	public double getZ(int id) {
		return vertsData[id * 3 + 2];
	}

	public void setVertData(int id, double x, double y, double z) {
		vertsData[id * 3] = x;
		vertsData[id * 3 + 1] = y;
		vertsData[id * 3 + 2] = z;
		vertexPositions[id].set(x, y, z);
	}

	public double getNX(int id) {
		return normalData[id * 3];
	}

	public double getNY(int id) {
		return normalData[id * 3 + 1];
	}

	public double getNZ(int id) {
		return normalData[id * 3 + 2];
	}

	public void update(MD5Frame frame) {
		for (MD5Vertex v : verts) {
			v.recalculatePosition(this, weights, frame);
		}
	}

	public double[] getTexCoords() {
		return uvData;
	}

	public double[] getNormals() {
		return normalData;
	}

	public double[] getPositions() {
		return vertsData;
	}

	public int[] getIndices() {
		return faceData;
	}

	public Vector3 getNormal(int index) {
		return bindPositionNormals[index];
	}

	public void setNormalData(int index, double x, double y, double z) {
		normalData[index * 3] = x;
		normalData[index * 3 + 1] = y;
		normalData[index * 3 + 2] = z;
	}
}
